package pnews.servlet;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jsoup.Jsoup;

import com.rometools.rome.feed.synd.SyndEnclosure;
import com.rometools.rome.feed.synd.SyndEntry;
import com.rometools.rome.feed.synd.SyndFeed;
import com.rometools.rome.io.FeedException;
import com.rometools.rome.io.SyndFeedInput;
import com.rometools.rome.io.XmlReader;

import pnews.Article;
import pnews.Category;
import pnews.Feed;

public class ArticleProvider {
        private static final Logger LOG = Logger.getLogger(ArticleProvider.class.getName());
        private final Map<Category, List<Article>> articlesByCategory = new HashMap<>();
        private final ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(2);
        private final Config config;
        
        public ArticleProvider(Config config) {
                this.config = config;
                for (Category cat: config.getCategories())
                        scheduler.scheduleAtFixedRate(new Refresher(cat), 2, 600, TimeUnit.SECONDS);
        }
        
        private static SyndFeed getSyndFeed(String u) throws IllegalArgumentException, FeedException, MalformedURLException, IOException {
                XmlReader r;
                
                r = new XmlReader(new URL(u));
                
                return new SyndFeedInput().build(r);                
        }
        
        private List<Article> getArticlesForUpdate(Category cat) {
                List<Article> result;
                
                synchronized (articlesByCategory) {
                        result = articlesByCategory.get(cat);
                        if (result == null) {
                                result = new ArrayList<>();
                                articlesByCategory.put(cat, result);
                        }
                        return result;
                }                
        }
        
        private boolean exists(String articleLink, List<Article> articles) {
                synchronized (articles) {
                        for (Article a: articles)
                                if (a.link.equals(articleLink))
                                        return true;
                }
                return false;
        }
        
        private static Article toArticle(String link, SyndEntry entry, SyndFeed feed) {
                String desc, title, thumbnail, feedTitle, str;
                Date date;
                
                feedTitle = feed.getTitle();
                if (feedTitle != null) {
                        feedTitle = feedTitle.trim();
                }
                
                thumbnail = null;
                for (SyndEnclosure e: entry.getEnclosures()) {
                        if (e.getType().startsWith("image/"))
                                thumbnail = e.getUrl();    
                        break;
                }
                
                if (thumbnail == null && feed.getImage() != null)
                        thumbnail = feed.getImage().getUrl();
                             
                
                title = entry.getTitle().trim();
                
                if (entry.getDescription() != null) {
                        str = entry.getDescription().getValue();
                        desc = Jsoup.parse(str).text();
                } else {       
                        desc = null;
                        LOG.severe("No description for " + feedTitle + " - " + title);
                }
                
                date = entry.getPublishedDate();
                if (date == null)
                        date = entry.getUpdatedDate();
                if (date == null)
                        LOG.severe("The article " + feedTitle + " - " + title + " does not have a date");
                                     
                return new Article(link, title, desc, thumbnail, date, feedTitle);
        }
        
        private void addArticles(Category cat, SyndFeed feed) {
                String feedTitle;
                List<Article> articles;
                Article a;
                
                feedTitle = feed.getTitle().trim();
                
                LOG.info("addArticles " + cat.getLabel() + " " + feedTitle + " number of articles: " + feed.getEntries().size());
                
                for (SyndEntry entry: feed.getEntries()) {
                        String link = entry.getLink().trim();
                        articles = getArticlesForUpdate(cat);
                        if (exists(link, articles)) {
                                LOG.fine("addArticles " + link + " is already present");
                                continue ;
                        }
                        
                        a = ArticleStore.singleton.getArticle(link, ()->toArticle(link, entry, feed));
                        
                        synchronized (articles) {
                                articles.add(a);

                                Collections.sort(articles, new Comparator<Article>() {
                                        @Override
                                        public int compare(Article o1, Article o2) {
                                                if (o1.publicationDate == o2.publicationDate)
                                                        return 0;
                                                if (o1.publicationDate == null)
                                                        return 1;
                                                if (o2.publicationDate == null)
                                                        return -1;
                                                return o2.publicationDate.compareTo(o1.publicationDate);
                                        }
                                });
                        }
                }          
                
                LOG.info("addArticles done " + cat.getLabel());
        }
             
        private void retrieveArticles(Category cat) throws IllegalArgumentException, MalformedURLException, FeedException, IOException {
                List<Feed> feeds;
                
                feeds = config.getFeedsByCategory().get(cat);
                
                if (feeds != null)
                        for (Feed f: feeds)
                                try {
                                        addArticles(cat, getSyndFeed(f.getURL()));
                                } catch (Throwable e) {
                                        LOG.log(Level.SEVERE,
                                                "retrieveArticles failure " + cat.getLabel() + " " + f.toString(),
                                                e);
                                }
                else
                        LOG.severe("No feed for category " + cat);
        }
        
        /**
         * Returns a copy.
         */
        public List<Article> getArticles(Category cat)
                        throws IllegalArgumentException, MalformedURLException, FeedException, IOException {
                List<Article> articles;
                
                synchronized (articlesByCategory) {
                        articles = getArticlesForUpdate(cat);
                }
                
                synchronized (articles) {
                        return new ArrayList<>(articles);
                }
        }
        
        private class Refresher implements Runnable {
                private final Category category;
                
                public Refresher(Category category) {
                        this.category = category;
                }
                
                @Override
                public void run() {                       
                        LOG.info("refresher "+ category.getLabel());
                        
                        try {
                                retrieveArticles(category);
                        } catch (IllegalArgumentException | FeedException | IOException e) {
                                LOG.log(Level.SEVERE, "refresher failure", e);
                        }                        
                        
                        LOG.info("refresher "+ category.getLabel() + " done");
                }                
        }
}
